/* Global constants */
#define DBL float

#define POLY_NMAX 32    /* max #sides to a polygon; change if needed */
#define MAX_SIDES_PER_POLYGON 32
#define MAX_MESH_SIZE 256
#define MAX_U_STEPS 256
#define MAX_V_STEPS 256
#define EPSILON 1.0e-5
#define PI 3.1415926535897932384626
#define TWO_PI (2.0 * PI)
#define TOTAL_COLORS 16
#define TOTAL_VGA_SHADES 64
#define TOTAL_EGA_SHADES 16
#define rad_per_degree  0.01745329251994329576
#define degree_per_rad 57.29577951308232087679

/* Definitions for ray-tracing */
#define MAX_RAY_TRACE_DEPTHS 256

/* Definitions for Radial objects. */
#define Default_Radial_Count 180
#define Default_Sector_Length 6
#define Default_Sector_Count 32
#define Default_Radial_Display 0

/* Global type definitions */
typedef struct Vector_Struct VECTOR;
struct Vector_Struct { DBL x, y, z; };
typedef struct Texture_Struct TEXTURE;
typedef struct Method_Struct METHODS;
typedef struct Object_Struct OBJECT;
typedef struct Viewpoint_Unique_Params FOCUS;
typedef struct Viewpoint_Object VIEWPOINT;
typedef struct Bezier_Unique_Params BEZIER_PARAMS;
typedef struct Bezier_Object BEZIER;
typedef struct Object_Struct ELLIPSE;
typedef struct Object_Struct HYPERBOLOID;
typedef struct Object_Struct PARABOLOID;
typedef struct Object_Struct CONE;
typedef struct Object_Struct HYPERBOLIC_PARABOLOID;
typedef struct Object_Struct PLANE;
typedef struct Radial_Unique_Params RADIAL_PARAMS;
typedef struct Radial_Struct RADIAL;
typedef struct Torus_Unique_Params TORUS_PARAMS;
typedef struct Torus_Struct TORUS;
typedef struct Coil_Unique_Params COIL_PARAMS;
typedef struct Coil_Struct COIL;
typedef struct Squishy_Unique_Params SQUISHY_PARAMS;
typedef struct Squishy_Struct SQUISHY;
typedef struct Polygon_Unique_Params POLYGON_PARAMS;
typedef struct Polygon_Struct POLYGON;
typedef struct Orientation_Struct ORIENTATION;
typedef struct Parameter_Struct PARAMETERS;
typedef struct Ray_Struct RAY;
typedef struct Depth_List_Struct DEPTH_LIST;

#define CHUNK 32767
typedef struct {
   unsigned char red, green, blue;
   } color_type;
typedef color_type * Screen_Map_Type;
typedef unsigned int depth_type;
typedef depth_type * Depth_Map_Type;

#define Viewpoint_Type              0
#define Light_Source_Type           1
#define Bezier_Type                 2
#define Ellipse_Type                3
#define Hyperboloid_Type            4
#define Paraboloid_Type             5
#define Plane_Type                  6
#define Torus_Type                  7
#define Coil_Type                   8
#define Hyperbolic_Paraboloid_Type  9
#define Polygon_Type               10
#define Squishy_Type               11
#define Radial_Type                12
#define Cone_Type                  13

typedef void (*PARAMETRIC_DISPLAYER)(OBJECT *, TEXTURE *);
typedef void (*PARAMETRIC_EVALUATER)(OBJECT *, DBL *, VECTOR *);
typedef void (*PARAMETRIC_NORMALER)(OBJECT *, DBL *, VECTOR *);
typedef void (*PARAMETRIC_ROTATER)(OBJECT *, VECTOR *);
typedef void (*PARAMETRIC_TRANSLATER)(OBJECT *, VECTOR *);
typedef void (*PARAMETRIC_SCALER)(OBJECT *, VECTOR *);
typedef void (*PARAMETRIC_MODIFIER)(OBJECT *);
struct Method_Struct {
   PARAMETRIC_DISPLAYER Displayer;
   PARAMETRIC_EVALUATER Evaluater;
   PARAMETRIC_NORMALER Normaler;
   PARAMETRIC_ROTATER Rotater1, Rotater2;
   PARAMETRIC_TRANSLATER Translater;
   PARAMETRIC_SCALER Scaler;
   };
struct Orientation_Struct {
   VECTOR Origin;
   VECTOR Major_Axis;
   VECTOR Minor_Axis1;
   VECTOR Minor_Axis2;
   VECTOR Scale;
   };
struct Parameter_Struct {
   DBL Min_U, Max_U, Min_V, Max_V;
   int U_Steps, V_Steps;
   };
struct Texture_Struct {
   DBL Color[4]; /* Red, Green, Blue, Alpha (for now), spectral */
                 /* samples in the future. */
   DBL Phong;
   };
struct Bezier_Unique_Params {
   VECTOR (*Patch)[4][4];
   DBL Division_Epsilon;
   };
struct Bezier_Object {
   METHODS *Methods;
   int Object_Type;
   OBJECT *Next_Object;
   TEXTURE Texture;
   ORIENTATION Orientation;
   PARAMETERS *Parameters;
   BEZIER_PARAMS *Bezier_Params;
   };
struct Viewpoint_Unique_Params {
   VECTOR Look_At;
   DBL Distance, Field_Of_View, Aspect;
   };
struct Viewpoint_Object {
   METHODS *Methods;
   int Object_Type;
   OBJECT *Next_Object;
   TEXTURE Texture;
   ORIENTATION Orientation;
   PARAMETERS *Parameters;
   FOCUS *Focus;
   };
struct Torus_Unique_Params {
   DBL a, b, h, k;
   };
struct Torus_Struct {
   METHODS *Methods;
   int Object_Type;
   OBJECT *Next_Object;
   TEXTURE Texture;
   PARAMETERS *Parameters;
   ORIENTATION Orientation;
   TORUS_PARAMS *Torus_Params;
   };
struct Coil_Unique_Params {
   DBL a, b, c, h, k;
   };
struct Coil_Struct {
   METHODS *Methods;
   int Object_Type;
   OBJECT *Next_Object;
   TEXTURE Texture;
   ORIENTATION Orientation;
   PARAMETERS *Parameters;
   COIL_PARAMS *Coil_Params;
   };
struct Squishy_Unique_Params {
   DBL a, b, c;
   };
struct Squishy_Struct {
   METHODS *Methods;
   int Object_Type;
   OBJECT *Next_Object;
   TEXTURE Texture;
   ORIENTATION Orientation;
   PARAMETERS *Parameters;
   SQUISHY_PARAMS *Squishy_Params;
   };
struct Polygon_Unique_Params {
   int Vertex_Count, Close_Flag;
   VECTOR *Vertices;
   };
struct Polygon_Struct {
   METHODS *Methods;
   int Object_Type;
   OBJECT *Next_Object;
   TEXTURE Texture;
   ORIENTATION Orientation;
   PARAMETERS *Parameters;
   POLYGON_PARAMS *Polygon_Params;
   };
struct Radial_Unique_Params {
   DBL Sector_Length;
   int Display_Type; /* Draw edges or not? */
   unsigned char plat[Default_Radial_Count][Default_Sector_Count];
   };
struct Radial_Struct {
   METHODS *Methods;
   int Object_Type;
   OBJECT *Next_Object;
   TEXTURE Texture;
   ORIENTATION Orientation;
   PARAMETERS *Parameters;
   RADIAL_PARAMS *Radial_Params;
   };
struct Object_Struct {
   METHODS *Methods;            /* Object unique methods */
   int Object_Type;             /* Defines what type of object this is. */
   OBJECT *Next_Object;         /* Link to the next object */
   TEXTURE Texture;             /* Texture information */
   ORIENTATION Orientation;     /* Major and minor axes */
   PARAMETERS *Parameters;      /* (u, v) parameters for the object */
   void *Dummy;                 /* Object specific parameters */
   };
struct Ray_Struct {
   VECTOR Initial;
   VECTOR Direction;
   };
struct Depth_List_Struct {
   OBJECT *Object;
   DBL Depth, U, V;
   };

typedef struct {       /* A POLYGON VERTEX */
   DBL sx, sy, sz, sw; /* screen space position (sometimes homogenous) */
   DBL x, y, z;        /* world space position */
   DBL nx, ny, nz;     /* world space normal vector */
   } Poly_vert;
/* update poly.c if you change this structure */

typedef struct {        /* A POLYGON */
    int n;                     /* number of sides */
    int mask;                  /* interpolation mask for vertex elems */
    Poly_vert vert[POLY_NMAX]; /* vertices */
} Poly;
/*
 * mask is an interpolation mask whose kth bit
 * indicates whether the kth DBL in a Poly_vert
 * is relevant. For example, if the valid attributes are
 * sx, sy, and sz, then set:
 *   mask = POLY_MASK(sx) | POLY_MASK(sy) | POLY_MASK(sz);
 */

typedef struct {        /* A BOX (TYPICALLY IN SCREEN SPACE) */
   DBL x0, x1;  /* left and right */
   DBL y0, y1;  /* top and bottom */
   DBL z0, z1;  /* near and far */
} Poly_box;


typedef struct {        /* WINDOW: A DISCRETE 2-D RECTANGLE */
   int x0, y0;  /* xmin and ymin */
   int x1, y1;  /* xmax and ymax (inclusive) */
} Window;

#define POLY_MASK(elem) (1<<(&poly_dummy->elem-(DBL *)poly_dummy))
#define POLY_CLIP_OUT 0         /* polygon entirely outside box */
#define POLY_CLIP_PARTIAL 1     /* polygon partially inside */
#define POLY_CLIP_IN 2          /* polygon entirely inside box */

extern Poly_vert *poly_dummy;   /* used superficially by  POLY_MASK 
macro */

typedef struct {
   VECTOR Pos;   /* Position of the light */
   VECTOR Color; /* Color of the light */
   DBL Intensity;/* Brightness of the light */
   } Light_Source;

#define MATRIX Matrix
typedef DBL Matrix[4][4];
typedef struct Transformation_Struct {
   Matrix matrix;
   Matrix inverse;
   } TRANSFORMATION;


#define VAdd(a, b, c) {(a).x=(b).x+(c).x;(a).y=(b).y+(c).y;(a).z=(b).z+(c).z;}
#define VSub(a, b, c) {(a).x=(b).x-(c).x;(a).y=(b).y-(c).y;(a).z=(b).z-(c).z;}
#define VHalf(a, b, c) {(a).x=((b).x+(c).x)/2.0;\
                        (a).y=((b).y+(c).y)/2.0;\
                        (a).z=((b).z+(c).z)/2.0;}
#define VScale(a, b, k)      {(a).x=(b).x*(k);(a).y=(b).y*(k);(a).z=(b).z*(k);}
#define VDot(a, b, c)        {a=(b).x*(c).x+(b).y*(c).y+(b).z*(c).z;}
#define VSet(a, x0, y0, z0)  {(a).x=x0;(a).y=y0;(a).z=z0;}
#define VInverseScale(a,b,k) {(a).x=(b).x/(k);(a).y=(b).y/(k);(a).z=(b).z/(k);}
#define VCross(a,b,c)        {(a).x=(b).y*(c).z-(b).z*(c).y; \
                              (a).y=(b).z*(c).x-(b).x*(c).z; \
                              (a).z=(b).x*(c).y-(b).y*(c).x;}
#define VLength(a, b)        a=sqrt((b).x*(b).x+(b).y*(b).y+(b).z*(b).z);
#define VDistance(d, a, b)   d=sqrt(((b).x - (a).x) * ((b).x - (a).x) +\
                                    ((b).y - (a).y) * ((b).y - (a).y) +\
                                    ((b).z - (a).z) * ((b).z - (a).z));
#define VNormalize(a,b)      {VTemp=sqrt((b).x*(b).x+(b).y*(b).y+(b).z*(b).z);\
                              (a).x=(b).x/VTemp;\
                              (a).y=(b).y/VTemp;\
                              (a).z=(b).z/VTemp;}

#define Evaluate(x, a, b) {if (x!=NULL)(*((x)->Methods->Evaluater))(x, a, b);}
#define Normal(x, a, b)   {if (x!=NULL)(*((x)->Methods->Normaler))(x, a, b);}
#define Show(x, y)        {if (x!=NULL)(*((x)->Methods->Displayer))(x, y);}
#define Translate(x,y)    {if (x!=NULL)(*((x)->Methods->Translater))(x, y);}
#define Scale(x,y)        {if (x!=NULL)(*((x)->Methods->Scaler))(x, y);}
#define Rotate1(x,y)      {if (x!=NULL)(*((x)->Methods->Rotater1))(x, y);}
#define Rotate2(x,y)      {if (x!=NULL)(*((x)->Methods->Rotater2))(x, y);}

#define sign(x) ((x) > 0 ? 1:  ((x) == 0 ? 0:  (-1)))
#define convert(x,y)   {x = (x + offsetx);  y = (offsety - y);}
#define unconvert(x,y) {x = (x - offsetx);  y = (offsety - y);}

/* Forward declarations for object manipulaters */
/* Generic object manipulaters */
void Generic_Displayer(OBJECT *, TEXTURE *);
void Generic_Translater(OBJECT *, VECTOR *);
void Generic_Scaler(OBJECT *, VECTOR *);
void Generic_Rotater1(OBJECT *, VECTOR *);
void Generic_Rotater2(OBJECT *, VECTOR *);
void Generic_Clipper(OBJECT *, VECTOR *, VECTOR *, int *, DBL *);

/* Viewpoint manipulaters */
void Viewpoint_Rotater1(OBJECT *, VECTOR *);

/* Object unique manipulaters  */
void Bezier_Displayer(OBJECT *, TEXTURE *);
void Bezier_Evaluater(OBJECT *, DBL *, VECTOR *);
void Bezier_Normaler(OBJECT *, DBL *, VECTOR *);

void Ellipse_Evaluater(OBJECT *, DBL *, VECTOR *);
void Ellipse_Normaler(OBJECT *, DBL *, VECTOR *);

void Hyperboloid_Evaluater(OBJECT *, DBL *, VECTOR *);
void Hyperboloid_Normaler(OBJECT *, DBL *, VECTOR *);

void Paraboloid_Evaluater(OBJECT *, DBL *, VECTOR *);
void Paraboloid_Normaler(OBJECT *, DBL *, VECTOR *);

void Cone_Evaluater(OBJECT *, DBL *, VECTOR *);
void Cone_Normaler(OBJECT *, DBL *, VECTOR *);

void Hyperbolic_Paraboloid_Evaluater(OBJECT *, DBL *, VECTOR *);
void Hyperbolic_Paraboloid_Normaler(OBJECT *, DBL *, VECTOR *);

void Plane_Evaluater(OBJECT *, DBL *, VECTOR *);
void Plane_Normaler(OBJECT *, DBL *, VECTOR *);

void Torus_Evaluater(OBJECT *, DBL *, VECTOR *);
void Torus_Normaler(OBJECT *, DBL *, VECTOR *);

void Coil_Evaluater(OBJECT *, DBL *, VECTOR *);
void Coil_Normaler(OBJECT *, DBL *, VECTOR *);

void Polygon_Displayer(OBJECT *, TEXTURE *);
void Polygon_Evaluater(OBJECT *, DBL *, VECTOR *);
void Polygon_Normaler(OBJECT *, DBL *, VECTOR *);

void Squishy_Evaluater(OBJECT *, DBL *, VECTOR *);
void Squishy_Normaler(OBJECT *, DBL *, VECTOR *);

void Radial_Evaluater(OBJECT *, DBL *, VECTOR *);
void Radial_Displayer(OBJECT *, TEXTURE *);
void Radial_Normaler(OBJECT *, DBL *, VECTOR *);

/* Various ray/line/plane routines */
int common_plane(VECTOR *, VECTOR *, VECTOR *, VECTOR *, VECTOR *, DBL *);
void project_points(VIEWPOINT *, int, VECTOR *, DBL *, DBL *, DBL *);
int polygon_normal(int, VECTOR *, VECTOR *, DBL *);
void clip_line(TEXTURE *, VECTOR *, VECTOR *);
int find_basis_values(VECTOR *, VECTOR *, VECTOR *, DBL *, DBL *);
int ray_parallelogram_intersect(int, VECTOR *, RAY *, DBL *, DBL *, DBL *);
int determine_plane_equations(int, VECTOR *, VECTOR *, DBL *, DBL *, DBL *,
                              VECTOR *, VECTOR *);

/* Math support routines */
DBL radians_to_degrees(DBL);
DBL degrees_to_radians(DBL);
void rotate_points(VECTOR *, int, VECTOR *, VECTOR *);
void determine_reference_angles(VECTOR *, VECTOR *);
DBL point_plane_distance(/* VECTOR *, VECTOR *, DBL * */);
void Normalize_View(TRANSFORMATION *WS);

/* Bugs in Turbo C cause these forward declarations to result in a
   "type mismatch" error when the function itself is compiled. */

/* int intersection_point(VECTOR *, DBL *, RAY *, VECTOR *, DBL); */


/* Display routines */
void initialize_graphics();
typedef int (*depth_proc)(DBL *, int, int);
void fill_poly(DBL *, TEXTURE *, int, int *, int *, depth_proc);
void fill_rectangle(DBL *, TEXTURE *, VECTOR *, VECTOR *, VECTOR *, VECTOR *);
DBL surface_test(VECTOR *, VECTOR *, VECTOR *);
void show_polygon(TEXTURE *, int, int, DBL *, DBL *);
void show_rectangle(TEXTURE *, VECTOR *, VECTOR *, VECTOR *, VECTOR *);
void put_depth(int, int, depth_type);
depth_type get_depth(int, int);
int depth_checker(DBL *, int, int);
void fill_rectangle(DBL *, TEXTURE *, VECTOR *, VECTOR *, VECTOR *, VECTOR *);
void drawRect(int, int, int, int, int);
void drawline(int, int, int, int, TEXTURE *);
int get_indexed_color(VECTOR);
void pallete_init(void);
DBL diffuse_shade(DBL *, VECTOR *);
void reset_pallette();
void put_color(int, int, DBL, DBL, DBL);
void TargaOpen(char *, int, int);
void TargaWrite();
void IndexedWrite();
void TargaClose();
void screen_dump();
#ifndef TURBOC
void line(int, int, int, int);
#endif

/* Viewpoint methods */
void reset_viewpoint();

/* Creation, modification, and I/O routines. */
OBJECT * create_object(int);
void wire_read_information(OBJECT *, FILE *);
void wire_write_information(OBJECT *, FILE *);
void DKB_Dump(OBJECT *, FILE *, int);
void get_orientation(OBJECT *);
void get_colors(OBJECT *);
void get_parameters(OBJECT *);
void get_unique_params(OBJECT *);

/* Prototypes for functions defined in matrices.c */
#define PARAMS(x) x
TRANSFORMATION *Get_Transformation PARAMS((void));
void MZero PARAMS((MATRIX *result));
void MIdentity PARAMS((MATRIX *result));
void MTimes PARAMS((MATRIX *result, MATRIX *matrix1, MATRIX *matrix2));
void MAdd PARAMS((MATRIX *result, MATRIX *matrix1, MATRIX *matrix2));
void MSub PARAMS((MATRIX *result, MATRIX *matrix1, MATRIX *matrix2));
void MScale PARAMS((MATRIX *result, MATRIX *matrix1, DBL amount));
void MTranspose PARAMS((MATRIX *result, MATRIX *matrix1));
void MTransformVector PARAMS((VECTOR *result, VECTOR *vector, TRANSFORMATION *transformation));
void MInverseTransformVector PARAMS((VECTOR *result, VECTOR *vector, TRANSFORMATION *transformation));
void Get_Scaling_Transformation PARAMS((TRANSFORMATION *result, VECTOR *vector));
void Get_Inversion_Transformation PARAMS((TRANSFORMATION *result));
void Get_Translation_Transformation PARAMS((TRANSFORMATION *transformation, VECTOR *vector));
void Get_Rotation_Transformation PARAMS((TRANSFORMATION *transformation, VECTOR *vector));
void Get_Look_At_Transformation PARAMS((TRANSFORMATION *transformation, VECTOR *Look_At, VECTOR *Up, VECTOR *Right));
void Get_Perspective_Transformation PARAMS((TRANSFORMATION *transformation, DBL dist));
void Compose_Transformations PARAMS((TRANSFORMATION *Original_Transformation, TRANSFORMATION *New_Transformation));
void Get_Perspective_Transformation PARAMS((TRANSFORMATION *, DBL));
